package milab.lab04.DoctorAgent;

import jade.core.Agent;
import jade.core.behaviours.*;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;

import javax.xml.bind.*;
import java.util.*;
import java.io.File;
import java.lang.Math;

import milab.*;
import milab.lab04.DoctorAgent.language01.*;
import milab.lab04.DoctorAgent.response.*;
import milab.lab04.DoctorAgent.DoctorAgentGui;
import milab.lab04.PatientAgent.language01.*;

// Bayes-Network reasoning JADE agent (c) Kovcs Dniel Lszl [dkovacs@mit.bme.hu]
public class DoctorAgent extends Agent {

	// Load the Bayes-Network DLL (Dynamic Link Library) from java.library.path, i.e. %JAVA_HOME%\bin directory
	static{ System.loadLibrary("agentBN_jni"); }

	// Path of the BNet-generated XML file describing the Bayes-Network of this DoctorAgent
	private String bayesNetworkFilePath;
	// Contains the filepath of the file holding the name of main parameter variable (mpv) and the symptoms 
	private String parametersFilePath;

	// Target parameter variable (the target of the Doctors' inference) - The Main Question! :)
	private String targetID;
	// Contains the names of parameter variables describing the PatientAgent's symptoms
	public String[] symptomIDs;
	// Contains a boolean whether a symptom is selected on the GUI to be used as evidence in the Bayes Network
	public boolean[] symptomIDsSelection;
	
  	// The GUI by means of which the user can check symptoms to be used as evidence
  	private DoctorAgentGui myGui;

	// Put agent initializations here
	protected void setup() {

		// Register the bayes-diagnostic service in the yellow pages
		DFAgentDescription dfd = new DFAgentDescription();
		dfd.setName(getAID());
		ServiceDescription sd = new ServiceDescription();
		sd.setType("bayes-diagnostic-service");
		sd.setName("bayes-diagnostic");
		dfd.addServices(sd);

		try {

			DFService.register(this, dfd);

		} catch (FIPAException fe) {

			fe.printStackTrace();

		}

		// Get Bayes-Network XML file's and Parameter Description file's filePath as start-up arguments
		Object[] args = getArguments();
    
		if (args != null && args.length > 1) {

			bayesNetworkFilePath = (String)args[0];
		    parametersFilePath = (String)args[1];
		    
		    File bayesNetworkFile = new File(bayesNetworkFilePath);
		    File parametersFile = new File(parametersFilePath);

		    if (bayesNetworkFile.isFile() && parametersFile.isFile()) {

		    	System.out.println("Bayes Network of agent " + getLocalName() + " is in: " + bayesNetworkFilePath);
		    	System.out.println("Parameter description file is: " + parametersFilePath);

		    	// Now read the parameter variables' name and the MPVi from the specified file
		    	String[] mpvAndSymptomIDs = DataHandler.readFile2String(parametersFilePath).split("\t");

		    	if (mpvAndSymptomIDs.length > 1) {

		    		targetID			= mpvAndSymptomIDs[0];
		    		symptomIDs			= new String[mpvAndSymptomIDs.length-1];
		    		symptomIDsSelection = new boolean[mpvAndSymptomIDs.length-1];
		    		
		    		for (int i = 1; i < mpvAndSymptomIDs.length; i++) {
		    			
		    			symptomIDs[i-1]				= mpvAndSymptomIDs[i];
		    			symptomIDsSelection[i-1]	= true;					// Let's assume that every symptom is checked at the beginning
		    			
		    		}
		    		System.out.println("Parameter variables' names parsed successfully...");

		    		// Create and show the GUI
		    		myGui = new DoctorAgentGui(this);
		    		myGui.show();
		    		
					// Add the behaviour serving queries from PatientAgent agents
					addBehaviour(new SymptomsServer());

		    	} else {
			  		
					// Make the agent terminate
				    System.out.println("[PARAMETER FILE ERROR] The Parameter description file has to declare the name of the target parameter and at least one symptom. Terminating...");
				    doDelete();

			  	}
		    	
		    } else {
		  		
				// Make the agent terminate
			    System.out.println("[FILE ERROR] One of the specified files is non existing. Terminating...");
			    doDelete();

		  	}
		    
		} else {

			// Make the agent terminate
		    System.out.println("[FILE ERROR] Not enough start-up arguments specified! A Bayes Network description file and a Parameter description file must be specified. Terminating...");
		    doDelete();

		}
		
	}

	// Put agent clean-up operations here
	protected void takeDown() {
  	
		// Deregister from the yellow pages
		try {
    	
			DFService.deregister(this);
      
		} catch (FIPAException fe) {
    	
			fe.printStackTrace();
      
		}

	  	// Close the GUI
	  	myGui.dispose();
		
		// Printout a dismissal message
		System.out.println("Doctor-agent " + getAID().getName() + " terminating.");
    
	}

  	private class SymptomsServer extends CyclicBehaviour {

	  public void action() {

		MessageTemplate mt = MessageTemplate.MatchPerformative(ACLMessage.REQUEST);
		ACLMessage msg = myAgent.receive(mt);

	    if (msg != null) {

	    	// REQUEST Message received. Excavating the necessary values...
	    	JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller();		// Creating an UnMarshaller
		    
	    	JAXBElement<MessageType> messageElement = (JAXBElement<MessageType>)jaxbUnMarshaller.readXMLString(msg.getContent(), "/jade/src/milab/lab04/patientagent/language01/language01.xsd", "milab.lab04.PatientAgent.language01");	// Taking the ACLMessageContent String as an XML, validating it, and unmarshalling it into an appropriate Java Object.
		    MessageType message	= messageElement.getValue();				// Getting the Java Object representing the XML root element...
		    
		    String[] symptoms 	= message.getSymptoms().split("\\s");		// Getting the separate String values of symptoms...
		    String[] diagnoses 	= message.getDiagnoses().split("\\s");		// Getting the separate String values of diagnoses...
		    		    
		    // ************************************************************************************************
		    // MOST MR TEHT RENDELKEZSRE LLNAK A PCIENS LTAL TKLDTT INFORMCIK, GY KVETKEZHET
		    // AZ INFORMCIK FELDOLGOZSA: A BAYES-HLS KVETKEZTETS! A BERKEZETT INFORMCIK A KVETKEZK:
		    //
		    // (1) SYMPTOMS:
		    //		STRING-EK TMBJE
		    //		ELEMEI "0", "1", VAGY "x". HA A TMB i-EDIK ELEME
		    //			"0", AZ AZT JELENTI, HOGY AZ i-EDIK PARAMTER A PCIENS SZERINT HAMIS
		    //			"1", AZ AZT JELENTI, HOGY AZ i-EDIK PARAMTER A PCIENS SZERINT IGAZ
		    //			"x", AZ AZT JELENTI, HOGY AZ i-EDIK PARAMTER RTKT A PCIENS NEM TUDJA
		    //
		    // (2) DIAGNOSES:
		    //		STRING-EK TMBJE
		    //		ELEMEI "0..1 kztti vals szm", VAGY "x". HA A TMB i-EDIK ELEME
		    //			"0..1", AZ AZT JELENTI, HOGY A KIEMELT (ELS) PARAMTER AZ i-EDIK ORVOS SZERINT EKKORA VALSZNSGGEL IGAZ
		    //			"x", AZ AZT JELENTI, HOGY A KIEMELT (ELS) PARAMTER RTKT AZ i-EDIK ORVOS NEM TUDJA
		    // ************************************************************************************************

		    // DIAGNOSE AGGREGATION!!!
		    String aggregatedDiagnoses = aggregateDiagnoses(diagnoses);
		    
		    // BAYES NETWORK INFERENCE!!!
		    String myDiagnoseString = bayesNetworkReasoning(symptoms, aggregatedDiagnoses, bayesNetworkFilePath);
		    
		    milab.lab04.DoctorAgent.language01.ObjectFactory factory = new milab.lab04.DoctorAgent.language01.ObjectFactory();
		    JAXBElement<String> diagnoseElement = factory.createDiagnose(myDiagnoseString);	// Creating the Java Object corresponding to the XML object
		    JAXBMarshaller jaxbMarshaller = new JAXBMarshaller();																					// Creating a universal Marshaller

		    ACLMessage reply = msg.createReply();
		    reply.setContent(jaxbMarshaller.generateXMLString("UTF-8", diagnoseElement, "milab.lab04.DoctorAgent.language01"));						// Marshalling the Java Object into a String with appropriate XML syntax	        
		    reply.setPerformative(ACLMessage.INFORM);

		    myAgent.send(reply);

	    } else {

		    block();

	    }

	  } // End of method SymptomsServer.action()

	}  // End of inner class SymptomsServer

  	// Aggregation of known diagnoses
  	private String aggregateDiagnoses(String[] diagnoseStrings) {
  	// INPUT IS AN ARRAY OF STRINGS OF DOUBLES BETWEEN 0.0 AND 1.0, OR X-ES
  	// OUTPUT MUST BE A STRING: "0", "1", or "X" !!!!!!!!!!!
  		
  		String returnValue;
  		boolean onlyx = true;			// Let's be "positive", and suppose, that there is only "x" on the input
  		double aggregatedValue = 0;
  		
  		for (int i = 0; i < diagnoseStrings.length; i++) {

  	  		if (!diagnoseStrings[i].equals("x")) {	// If a doctor gave a prediction

  	  			onlyx = false;
  	  			aggregatedValue += Double.parseDouble(diagnoseStrings[i]);
  	  			
  	  		} // Else the diagnose is assumed to be "x"

  		} // End of FOR
  		
  		if (onlyx) {
  			
  			returnValue = "x";
  			
  		} else {
  			
  			if ((aggregatedValue / diagnoseStrings.length) < 0.5) {
  	  			
  				returnValue = "0";
  			
  			} else {
  			
  				returnValue = "1";
  			
  			}
  			
  		} // End of IF
  		
  		return returnValue;

  	} // End of method DoctorAgent.aggregateDiagnoses
  	
	// function accessing the native library
	private native String doAction(String request);

	// implementation of an inference session
	// load network, enter evidence, infer, free network (c) Millinghoffer Andrs [milli@mit.bme.hu]
	public String bayesNetworkReasoning(String[] symptoms, String aggregatedDiagnoses, String bayesNetworkFilePath){

		// load network
		String response = doAction(
				"<BNCL>\n" +
				"    <loadBNCall callId=\"1\" classId=\"myNetwork\" fileLocation=\"" + bayesNetworkFilePath + "\"/>\n" +
				"</BNCL>"
				);

		// clear evidence
		response = doAction(
				"<BNCL>\n" +
				"    <clearEvidenceCall callId=\"6\" classId=\"myNetwork\"/>\n" +
				"</BNCL>"
				);

		// enter evidence
		String request =	"<BNCL>\n" +
							"	<setEvidenceCall callId=\"3\" classId=\"myNetwork\">\n";

		for (int i = 0; i < symptomIDs.length; i++ ) {

			if (symptoms[i].equalsIgnoreCase("x") || !symptomIDsSelection[i]) {

				continue;

			} else {

				request += "		<node id=\"" + symptomIDs[i] + "\" evidence=\"" + (symptoms[i].equals("0") ? "1.0;0.0" : "0.0;1.0") + "\"/>\n";

			}

		}

		if(!aggregatedDiagnoses.equalsIgnoreCase("x")) {

			request += "		<node id=\"aggregatedDiagnoses\" evidence=\"" + (aggregatedDiagnoses.equals("0") ? "1.0;0.0" : "0.0;1.0") + "\"/>\n";

		}

		request +=	"	</setEvidenceCall>\n" +
					"</BNCL>";

		//System.out.println("*** " + request);
		
		doAction(request);

		// get marginal of target
		response = doAction(
				"<BNCL>\n" +
				"    <inferMarginalCall callId=\"4\" classId=\"myNetwork\">\n" +
				"        <node id=\"" + targetID + "\"/>\n" +
				"    </inferMarginalCall>\n" +
				"</BNCL>"
				);

		//System.out.println("**** " + response);
		
    	// Got the response - excavating the necessary value...
    	JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller();		// Creating an UnMarshaller
    	JAXBElement<BNCLType> BNCLElement = (JAXBElement<BNCLType>)jaxbUnMarshaller.readXMLString(response, "/jade/src/milab/lab04/doctoragent/response/response.xsd", "milab.lab04.DoctorAgent.response");	// Taking the response String as an XML, validating it, and unMarshalling it into an appropriate Java Object.
    	String[] marginal	= BNCLElement.getValue().getInferMarginalReturn().getNode().getMarginal().split(";");				// Getting the marginal XML-attribute's value from the node element of the inferMarginalReturn element from the BNCL root element...   

		// free network
		response = this.doAction(
				"<BNCL>\n" +
				"    <freeBNCall callId=\"7\" classId=\"myNetwork\"/>\n" +
				"</BNCL>"
				);

		return marginal[1];	// Return the probability of the target-node being true (marginal.length == 2)

	} // End of method DoctorAgent.bayesNetworkReasoning

} // End of class DoctorAgent