package milab.lab04.PatientAgent;

import jade.core.AID;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

// GUI for Patient JADE agent (c) Kovcs Dniel Lszl [dkovacs@mit.bme.hu]
class PatientAgentGui extends JFrame {

	private PatientAgent myAgent;

	public JButton newButton;
	public JButton diagButton;
	public JComboBox doctors;
	public JComboBox[] symptom_values;
	public JTextField diagnose_value;
	public String[] param_values = {"false", "true", "unknown"};
	public ComboBoxActionListener cbal = new ComboBoxActionListener();

	PatientAgentGui(PatientAgent a) {

		super(a.getLocalName());

		myAgent = a;
		
		// ---------------------------- NORTHERN PART OF THE FRAME ----------------------------
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(2, 1));
		
		// ---------------------------- PART 1 ----------------------------
		JPanel p1 = new JPanel();
		p1.setLayout(new GridLayout(1, 1));
		
		doctors = new JComboBox();
		doctors.setPrototypeDisplayValue("1235678901234567890123456789012");
		doctors.setEnabled(false);
		p1.add(doctors);		
		
		p1.setBorder(BorderFactory.createTitledBorder("Doctor(s)"));
		p.add(p1);		
		
		// ---------------------------- PART 2 ----------------------------
		JPanel p2 = new JPanel();
		p2.setLayout(new GridLayout(1, 2));
		
		JLabel lx = new JLabel("Diagnosis for " + myAgent.targetID);
		lx.setHorizontalAlignment(JLabel.CENTER);
		p2.add(lx);
		
		diagnose_value	= new JTextField("");
		diagnose_value.setEnabled(false);
		p2.add(diagnose_value);

		p2.setBorder(BorderFactory.createTitledBorder("My diagnosis"));
		p.add(p2);
		
		getContentPane().add(p, BorderLayout.NORTH);
		
		// ---------------------------- CENTRAL PART OF THE FRAME ----------------------------
		p = new JPanel();
		p.setLayout(new GridLayout(myAgent.symptomIDs.length + 1, 2));
		
		lx = new JLabel("Symptom name");
		lx.setHorizontalAlignment(JLabel.CENTER);
		p.add(lx);
		
		lx = new JLabel("Symptom value");
		lx.setHorizontalAlignment(JLabel.CENTER);
		p.add(lx);
		
		// Memory reservation...
		symptom_values	= new JComboBox[myAgent.symptomIDs.length];
		
		// Ready up the main JPanel in this JFrame
		for (int i = 0; i < myAgent.symptomIDs.length; i++) {
			
			lx = new JLabel(myAgent.symptomIDs[i]);
			lx.setHorizontalAlignment(JLabel.CENTER);
			p.add(lx);
			
			symptom_values[i] = new JComboBox(param_values);
			symptom_values[i].setSelectedIndex(2);
			symptom_values[i].setEnabled(false);
			p.add(symptom_values[i]);

		}
		
		p.setBorder(BorderFactory.createTitledBorder("My symptoms"));
		getContentPane().add(p, BorderLayout.CENTER);

		// ---------------------------- SOUTHERN PART OF THE FRAME ----------------------------
		p = new JPanel();
		p.setLayout(new GridLayout(1, 2));
		
		newButton = new JButton("New symptoms");
		newButton.addActionListener( new ActionListener() {

			public void actionPerformed(ActionEvent ev) {
				
				try {
					
					diagnose_value.setText("");
					doctors.removeActionListener(cbal);
					doctors.removeAllItems();
					doctors.setEnabled(false);
					myAgent.activateBehaviour("InitializeNewRound");
					
				} catch (Exception e) {
					
					JOptionPane.showMessageDialog(PatientAgentGui.this, "Invalid values. " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE); 
					
				}
				
			}
			
		});
		
		diagButton = new JButton("Diagnose (again)");
		diagButton.addActionListener( new ActionListener() {

			public void actionPerformed(ActionEvent ev) {
				
				try {
					
					doctors.removeActionListener(cbal);
					diagnose_value.setText("");
					
					for (int i = 0; i < symptom_values.length; i++) {

						symptom_values[i].setEnabled(false);
						
				    }
					
					newButton.setEnabled(false);
					diagButton.setEnabled(false);
					doctors.setEnabled(false);
					myAgent.activateBehaviour("RequestDiagnoses");
					
				} catch (Exception e) {
					
					JOptionPane.showMessageDialog(PatientAgentGui.this, "Invalid values. "+e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE); 
					
				}
				
			}
			
		});
		diagButton.setEnabled(false);
		
		p.add(newButton);
		p.add(diagButton);
		
		getContentPane().add(p, BorderLayout.SOUTH);
		
		// --------------------------------------------------
		// Make the agent terminate when the user closes 
		// the GUI using the button on the upper right corner	
		addWindowListener(new WindowAdapter() {
			
			public void windowClosing(WindowEvent e) {
				
				myAgent.doDelete();
				
			}
			
		});
		
		setResizable(false);
		
	} // End of PatientAgentGui's constructor
	
	public class ComboBoxActionListener implements ActionListener {

		public void actionPerformed(ActionEvent ev) {
			
			try {

				diagnose_value.setText(myAgent.mpvDiagnoses[doctors.getSelectedIndex()]);

			} catch (Exception e) {

				JOptionPane.showMessageDialog(PatientAgentGui.this, "Invalid values. " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE); 

			}

		}

	} // End of inner class ComboBoxActionListener
	
	public void show() {
		
		pack();
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		int centerX = (int)screenSize.getWidth() / 2;
		int centerY = (int)screenSize.getHeight() / 2;
		setLocation(centerX - getWidth() / 2, centerY - getHeight() / 2);
		super.show();
		
	} // End of method PatientAgentGui.show()
	
} // End of class PatientAgentGui
